/*
    This file is part of the oolaboola digital dj program

    Copyright (C) 1999  Eric S. Tiedemann

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*

  in GtkObjectFromPyGtkWidget(), confirm that it's an instance of GtkWidget
  use something instead of calloc()
  clean up and check types well
  free resources on init errors
  get issues to james
  make more objecty
  __del__ methods

*/

#include <python1.5/Python.h>
#include <gtk/gtk.h>
#include "gtkdial.h"
#include "ivscale.h"

/* euch */
typedef struct {
  PyObject_HEAD
  GtkObject *obj;
} PyGtk_Object;

typedef PyObject *(*PGN)(GtkObject *go);
static PGN pgn;
static PyTypeObject *pgot;

struct dial_data
{
  PyObject *cb;
  int sigid;
  GtkObject *adj;
};

int
invoke_cb_on_formatted_adj_value(PyObject *cb, char *fmt, GtkObject *adj)
{
  PyObject *o;
  char buf[1024];

  snprintf(buf, sizeof(buf), fmt, GTK_ADJUSTMENT(adj)->value);
  if ((o = PyObject_CallFunction(cb, "(s)", buf)))
    {
      Py_DECREF(o);
      return 1;
    }
  else
    return 0;
}

void
call_dial_cb(GtkObject *adj, struct dial_data *dcb)
{
  PyObject *o;

  if ((o = PyObject_CallFunction(dcb->cb, "(d)", GTK_ADJUSTMENT(adj)->value)))
    Py_DECREF(o);
}

/* !!! need to clean up nicely on failure */
static PyObject *
oolalow_dial_new(PyObject *self, PyObject *args)
{
  PyObject *cb;
  double val, min, max, inc, page, digits;
  struct dial_data *dcb;
  GtkObject *adj;
  GtkWidget *d;

  if (!PyArg_ParseTuple(args, "ddddddO;oolalow.dial", 
			&val, &min, &max, &inc, &page, &digits,
			&cb))
    return 0;

  if (!PyCallable_Check(cb))
    {
      PyErr_SetString(PyExc_TypeError, "callback argument isn't callable!");
      return 0;
    }

  Py_INCREF(cb);

  /* need to set exception and/or use python provided allocater */
  dcb = calloc(1, sizeof(struct dial_data));

  if (!dcb)
    return 0;

  dcb->cb = cb;

  dcb->adj = gtk_adjustment_new(val, min, max, inc, page, 0);

  if (!dcb->adj) return 0;

  /* !!! check return value */
  dcb->sigid = gtk_signal_connect(dcb->adj, 
				  "value_changed", 
				  call_dial_cb, 
				  dcb);

  d = gtk_dial_new(GTK_ADJUSTMENT(dcb->adj));

  if (!d) return 0;

  gtk_object_set_data(GTK_OBJECT(d), "oolalow", dcb);

  return pgn(GTK_OBJECT(d));
}

/* set adj's value to v without triggering the signal handler
   registered with id sigid */
static void
diddle_adjustment(GtkObject *adj, int sigid, double v)
{
  gtk_signal_handler_block(adj, sigid);
  gtk_adjustment_set_value(GTK_ADJUSTMENT(adj), v);
  gtk_signal_handler_unblock(adj, sigid);
  gtk_signal_emit_by_name(adj, "changed");
}

static GtkObject *
GtkObjectFromPyGtkWidget(PyObject *w)
{
  PyObject *_o = PyObject_GetAttrString(w, "_o");

  if (! _o)
    return 0;

  if (_o->ob_type != pgot)
    return 0;

  return ((PyGtk_Object *) _o)->obj;
}

static PyObject *
oolalow_dial_set(PyObject *self, PyObject *args)
{
  PyObject *o;
  double v;
  PyObject *_o;
  GtkObject *adj;
  struct dial_data *dcb;
  GtkObject *go;

  if (!PyArg_ParseTuple(args, "O!d;oolalow.dial_set", 
			&PyInstance_Type, &o, &v))
    return 0;

  if (!(go = GtkObjectFromPyGtkWidget(o)))
    return 0;

  if (!(dcb = gtk_object_get_data(GTK_OBJECT(go), "oolalow")))
    return 0;

  adj = dcb->adj;

  diddle_adjustment(adj, dcb->sigid, v);

  call_dial_cb(adj, dcb);

  Py_INCREF(Py_None);
  return Py_None;
}

struct ovscale_data
{
  PyObject *cb;
  int sigid;
  GtkObject *adj;
  char *fmt, *pause, *unpause;
  int pausedp;
};

int
call_ovscale_cb(GtkObject *adj, struct ovscale_data *ovd)
{
  if (!ovd->pausedp)
    return invoke_cb_on_formatted_adj_value(ovd->cb, ovd->fmt, adj);
  else
    return 1;
}

void
call_ovscale_cb1(GtkObject *adj, struct ovscale_data *ovd)
{
  call_ovscale_cb(adj, ovd);
}

/* !!! need to clean up nicely on failure */
static PyObject *
oolalow_ovscale_new(PyObject *self, PyObject *args)
{
  PyObject *cb;
  PyObject *fmto, *pauseo, *unpauseo;
  double val, min, max, inc, page;
  int digits;
  struct ovscale_data *ovd;
  GtkWidget *v;

  if (!PyArg_ParseTuple(args, "dddddiOO!O!O!;oolalow.ovscale", 
			&val, &min, &max, &inc, &page, &digits,
			&cb,
			&PyString_Type, &fmto, 
			&PyString_Type, &pauseo, 
			&PyString_Type, &unpauseo))
    return 0;

  if (!PyCallable_Check(cb))
    {
      PyErr_SetString(PyExc_TypeError, "callback argument isn't callable!");
      return 0;
    }

  ovd = calloc(1, sizeof(struct ovscale_data));

  if (!ovd)
    return 0;

  Py_INCREF(fmto);
  Py_INCREF(pauseo);
  Py_INCREF(unpauseo);
  Py_INCREF(cb);

  ovd->fmt = PyString_AsString(fmto);
  ovd->pause = PyString_AsString(pauseo);
  ovd->unpause = PyString_AsString(unpauseo);
  ovd->cb = cb;


  ovd->adj = gtk_adjustment_new(val, min, max, inc, page, 0);
  if (!ovd->adj) return 0;

  /* !!! check return value */
  ovd->sigid = gtk_signal_connect(ovd->adj, 
				  "value_changed", 
				  call_ovscale_cb1, 
				  ovd);

  v = ivscale_new(GTK_ADJUSTMENT(ovd->adj));

  ivscale_set_digits(IVSCALE(v), digits);

  gtk_object_set_data(GTK_OBJECT(v), "oolalow", ovd);

  if (!v) return 0;

  return pgn(GTK_OBJECT(v));
}

static PyObject *
ovscale_access(PyObject *args, 
	       char *pfmt,
	       int (*f)(struct ovscale_data *, double))
{
  PyObject *o;
  double v = 0;
  PyObject *_o;
  GtkObject *adj;
  struct ovscale_data *ovd;
  GtkObject *go;

  if (!PyArg_ParseTuple(args, pfmt, &PyInstance_Type, &o, &v))
    return 0;

  if (!(go = GtkObjectFromPyGtkWidget(o)))
    return 0;

  if (!(ovd = gtk_object_get_data(GTK_OBJECT(go), "oolalow")))
    return 0;

  if (f(ovd, v))
    {
      Py_INCREF(Py_None);
      return Py_None;
    }
  else
    return 0;
}

int
ovscale_set(struct ovscale_data *ovd, double v)
{
  diddle_adjustment(ovd->adj, ovd->sigid, v);
  return call_ovscale_cb(ovd->adj, ovd);
}

static PyObject *
oolalow_ovscale_set(PyObject *self, PyObject *args)
{
  return(ovscale_access(args, "O!d;oolalow.ovscale_set", ovscale_set));
}

int
ovscale_pause(struct ovscale_data *ovd, double v)
{
  ovd->pausedp = 1;

  return invoke_cb_on_formatted_adj_value(ovd->cb, ovd->pause, ovd->adj);
}

static PyObject *
oolalow_ovscale_pause(PyObject *self, PyObject *args)
{
  return(ovscale_access(args, "O!;oolalow.ovscale_pause", ovscale_pause));
}

int
ovscale_unpause(struct ovscale_data *ovd, double v)
{
  ovd->pausedp = 0;

  return invoke_cb_on_formatted_adj_value(ovd->cb, ovd->unpause, ovd->adj);
}

static PyObject *
oolalow_ovscale_unpause(PyObject *self, PyObject *args)
{
  return(ovscale_access(args, "O!;oolalow.ovscale_unpause", ovscale_unpause));
}

static PyMethodDef oolalowMethods[] =
{
  {"ovscale", oolalow_ovscale_new, METH_VARARGS},
  {"ovscale_set", oolalow_ovscale_set, METH_VARARGS},
  {"ovscale_pause", oolalow_ovscale_pause, METH_VARARGS},
  {"ovscale_unpause", oolalow_ovscale_unpause, METH_VARARGS},
  {"dial", oolalow_dial_new, METH_VARARGS},
  {"dial_set", oolalow_dial_set, METH_VARARGS},
  {0, 0, 0}
};

void
initoolalow()
{
  PyObject *_gtkmod, *_gtkmodd, *_private, *pgncobj;
  PyObject *gtk;
  PyObject *gtkd;

  if (!(_gtkmod = PyImport_ImportModule("_gtk")))
    Py_FatalError("couldn't import _gtk");

  if (!(_gtkmodd = PyModule_GetDict(_gtkmod)))
    Py_FatalError("couldn't access _gtk's dictionary!");

  Py_DECREF(_gtkmod);

  if (!(pgot = 
	(PyTypeObject *) PyDict_GetItemString(_gtkmodd, "GtkObjectType")))
    Py_FatalError("couldn't access _gtk.GtkObjectType");

  Py_INCREF(pgot);

  if (!(_private = PyDict_GetItemString(_gtkmodd, "_private")))
    Py_FatalError("couldn't access _gtk._private");

  if (!(pgncobj = PyDict_GetItemString(_private, "PyGtk_New")))
    Py_FatalError("couldn't access _gtk._private.PyGtk_New");

  if (! PyCObject_Check(pgncobj))
    Py_FatalError("gtk._private['PyGtk_New'] isn't a CObject!");

  pgn = (PGN) PyCObject_AsVoidPtr(pgncobj);

  Py_DECREF(pgncobj);

  Py_InitModule("oolalow", oolalowMethods);

  if (PyErr_Occurred())
    Py_FatalError("can't initialize module oolalow");
}
