/*
    This file is a part of the estlib C++ library.
    Copyright (C) 1999  Eric S. Tiedemann

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <queue>
extern "C"
{
#include <pthread.h>
}

#include "mutex.h"

namespace estlib
{
template <class element>
class ServiceQueue
{
  enum {INHERIT_SCHEDULING = 1};
  pthread_mutex_t m;
  pthread_cond_t c;
  pthread_cond_t empty;
  pthread_t t;
  std::queue<element> q;
  static void *task(void *);
  //void (*cb)(FILE *, void *);
  //void *data;
  //struct timespec interval, next;
  //set_periodic(void (*cb)(FILE *, void *), 
  //             void *data, 
  //             struct timespec *interval);

protected:

  virtual void service(element e) = 0;

public:

  ServiceQueue(unsigned opts = 0);
  ~ServiceQueue();
  void push(const element &);
  void wait();
};

/*
  TODO
  check for errors..throw exceptions
  handle periodic tasks
  figger where all this should live in library dist
*/

template <class element>
void
ServiceQueue<element>::push(const element &e)
{
  estlib::mutex x(&m);
  
  q.push(e);
  pthread_cond_signal(&c);
}

template <class element>
ServiceQueue<element>::~ServiceQueue()
{
  pthread_cancel(t);
  pthread_mutex_destroy(&m);
  pthread_cond_destroy(&c);
  pthread_cond_destroy(&empty);
}

template <class element>
ServiceQueue<element>::ServiceQueue(unsigned opts)
{
  pthread_mutex_init(&m, 0);
  pthread_cond_init(&c, 0);
  pthread_cond_init(&empty, 0);

  if (opts & INHERIT_SCHEDULING)
    {
      pthread_attr_t att;
      pthread_attr_init(&att);
      pthread_attr_setinheritsched(&att, PTHREAD_INHERIT_SCHED);
      pthread_create(&t, &att, task, this);
      pthread_attr_destroy(&att);
    }
  else
    pthread_create(&t, 0, task, this);
}

template <class element>
void
ServiceQueue<element>::wait()
{
  estlib::mutex x(&m);
  while (q.size() > 0)
    pthread_cond_wait(&empty, &m);
}

template <class element>
void *
ServiceQueue<element>::task(void *v)
{
  ServiceQueue *c = static_cast<ServiceQueue *>(v);
  
  pthread_mutex_lock(&c->m);

  while (true)
    {
      while (c->q.size() == 0)
	{
	  pthread_cond_wait(&c->c, &c->m);
	  // handle periodic task here
	}
      
      while (c->q.size() > 0)
	{
	  element e = c->q.front();
	  c->q.pop();
	  pthread_mutex_unlock(&c->m);
	  c->service(e);
	  pthread_mutex_lock(&c->m);
	}

      pthread_cond_broadcast(&c->empty);
    }
}

}
